/*
****************************************************************************
****************************************************************************
**                                                                        **
**                XXXX     X    X         X XXXXXX X                      **
**               X    X   X X   X    X    X X      X   X                  **
**               X       X   X  X   X X   X X      X  X                   **
**               X   XX XXXXXXX X  X   X  X XXXX   X X                    **
**               X    X X     X X X     X X X      XX                     **
**               X    X X     X XX       XX X      X                      **
**                XXXX  X     X X         X XXXXXX XXXXXX                 **
**                                                                        **
**                                                                        **
**                       Copyright (C) A. P. H.                           **
**                           BIALYSTOK 2007                               **
**                                                                        **
**                      MODULE DHCP Protocol Service                      **
**                                                                        **
**                     Mikrokontroler ARM: LPC2378                        **
**                                                                        **
**                     Wersja z dnia : 2007.03.22                         **
**                     Napisal : Andrzej Pawluczuk                        **
**                                                                        **
****************************************************************************
****************************************************************************
*/
//rfc1531
#include "dhcpservice.h"
#include "systime.h"
#include "udpservice.h"
#include "types.h"
#include "ipservice.h"
#include "frameutil.h"
#include "uart0.h"
#include "trace.h"


#define TickPerSecond                  1000

#define DHCPClientPortNo               68
#define DHCPServerPortNo               67
#define DHCP_MagicCookie               0x63825363

#define DHCP_BootRequest               1
#define DHCP_BootReply                 2

/*      Options codes  */
#define DHCP_PadCode                   0
#define DHCP_SubNetMaskCode            1
#define DHCP_RouterCode                3
#define DHCP_HostNameCode              12
#define DHCP_RequestedIPCode           50
#define DHCP_LeaseTimeCode             51
#define DHCP_MessageTypeCode           53
#define DHCP_ServerIDCode              54
#define DHCP_ParameterRequestListeCode 55
#define DHCP_ClientCode                61
#define DHCP_EndCode                   255

/*  DHCP message types.  */
#define DHCP_DHCPDISCOVER              1
#define DHCP_DHCPOFFER                 2
#define DHCP_DHCPREQUEST               3
#define DHCP_DHCPDECLINE               4
#define DHCP_DHCPACK                   5
#define DHCP_DHCPNACK                  6
#define DHCP_DHCPRELEASE               7
#define DHCP_DHCPINFORM                8

#define DHCP_HEADER_SIZE               44
#define DHCP_SERVERNAME_SIZE           64
#define DHCP_STARTFILE_SIZE            128
#define DHCP_DISCOVER_OPTIONS_SIZE     8
#define DHCP_REQUEST_OPTIONS_SIZE      24
#define DHCP_MINIMUM_PACKET_SIZE ( DHCP_HEADER_SIZE + DHCP_SERVERNAME_SIZE + DHCP_STARTFILE_SIZE + DHCP_DISCOVER_OPTIONS_SIZE )

//#define DHCP_MIN_TIMEOUT 4
#define DHCP_REQUEST_RETRIES           4
#define DHCP_INFINITE_DURATION         0xffffffff

#define DHCP_TimeOut                   ( 5 * TickPerSecond )

static DHCPLayerInstanceRecT LocalDHCPLayerInstance ;

static USHORT SendDHCPRequest ( PtrDHCPLayerInstanceRecT /* DHCPInstance */ ) ;


static void DHCPLeaveTimeOutNotify ( void * CallParam )
{
  PtrDHCPLayerInstanceRecT DHCPInstance ;
  PtrIPLayerInstanceRecT IPLayerInstance ;
  PtrUDPLayerInstanceRecT UDPLayerInstance ;
  /*-------------------------------------------------------------------------*/
  DHCPInstance = ( PtrDHCPLayerInstanceRecT ) CallParam ;
  UDPLayerInstance = DHCPInstance -> UDPLayerInstance ;
  IPLayerInstance = UDPLayerInstance -> IPLayerInstance ;
  if ( DHCPInstance -> TimeOutService )
  {
    DetachTimer ( DHCPInstance -> TimeOutService ) ;
    DHCPInstance -> TimeOutService = NIL ;
  } /* if */ ;
  SetIPAdresses ( IPLayerInstance , 0 , 0 , 0 ) ;
  DHCPClientRenew ( DHCPInstance ) ;
} /* DHCPLeaveTimeOutNotify */


static ULONG MagicID ( PtrUDPLayerInstanceRecT UDPLayerInstance )
{
  ULONG Status ;
  PtrIPLayerInstanceRecT IPLayerInstance ;
  EMACInstanceRecT * InterfInstance ;
  /*-------------------------------------------------------------------------*/
  IPLayerInstance = UDPLayerInstance -> IPLayerInstance ;
  InterfInstance = IPLayerInstance -> InterfaceInstance ;
  Status = ( ( ULONG ) InterfInstance -> InterfaceMACAddress [ 2 ] << 24 ) |
           ( ( ULONG ) InterfInstance -> InterfaceMACAddress [ 3 ] << 16 ) |
           ( ( ULONG ) InterfInstance -> InterfaceMACAddress [ 4 ] << 8 ) |
           ( ( ULONG ) InterfInstance -> InterfaceMACAddress [ 5 ] ) ;
  Status ^= GetRandom ( ) ;
  return ( Status ) ;
} /* MagicID */


static ULONG LocGetULONG ( UCHAR * DataBuffer )
{
  ULONG Temp ;
  /*-------------------------------------------------------------------------*/
  Temp = ( ULONG ) ( * DataBuffer ++ ) ;
  Temp = ( Temp << 8 ) | ( ULONG ) ( * DataBuffer ++ ) ;
  Temp = ( Temp << 8 ) | ( ULONG ) ( * DataBuffer ++ ) ;
  Temp = ( Temp << 8 ) | ( ULONG ) ( * DataBuffer ++ ) ;
  return ( Temp ) ;
} /* LocGetULONG */


static void DHCPDataRecvNotify ( PtrUDPSocketDescriptionRecT UDPSocket ,
                                 ULONG                       SenderIPAddress ,
                                 USHORT                      SenderPortNo ,
                                 ULONG                       DestIPAddress ,
                                 USHORT                      DestPortNo ,
                                 void *                      UDPDataBuffer ,
                                 USHORT                      UDPDataSize )
{
  PtrDHCPLayerInstanceRecT DHCPInstance ;
  PtrIPLayerInstanceRecT IPLayerInstance ;
  PtrUDPLayerInstanceRecT UDPLayerInstance ;
  UCHAR * DataBuffer ;
  UCHAR Operation ;
  UCHAR Option ;
  UCHAR OptionSize ;
  UCHAR MessageType ;
  USHORT Reply ;
  ULONG XID ;
  ULONG ClientIP ;
  ULONG Cookie ;
  ULONG Server_ID ;
  ULONG SubNet ;
  ULONG GateWay ;
  ULONG Duration ;
  /*-------------------------------------------------------------------------*/
traceDHCP(UDPDataBuffer , UDPDataSize);
  DHCPInstance = & LocalDHCPLayerInstance ;
  Reply = FALSE ;
  MessageType = 0 ;
  Server_ID = 0 ;
  SubNet = 0 ;
  GateWay = 0 ;
  Duration = DHCP_INFINITE_DURATION ;
  DataBuffer = ( UCHAR * ) UDPDataBuffer ;
  if ( ( DHCPInstance -> Command ) && ( UDPDataSize >= DHCP_MINIMUM_PACKET_SIZE ) )
  {
    Operation = * DataBuffer ++ ; UDPDataSize -- ;
    DataBuffer += 3 ; UDPDataSize -= 3 ;
    XID = LocGetULONG ( DataBuffer ) ;
    DataBuffer += 4 ; UDPDataSize -= 4 ;
    DataBuffer += 8 ; UDPDataSize -= 8 ;
    ClientIP = LocGetULONG ( DataBuffer ) ;
    DataBuffer += 4 ; UDPDataSize -= 4 ;
    DataBuffer += 24 ; UDPDataSize -= 24 ;
    if ( ( Operation == DHCP_BootReply ) && ( XID == DHCPInstance -> XID ) && ClientIP )
    {
      DataBuffer += ( DHCP_SERVERNAME_SIZE + DHCP_STARTFILE_SIZE ) ;
      Cookie = LocGetULONG ( DataBuffer ) ;
      DataBuffer += 4 ; UDPDataSize -= 4 ;
      if ( Cookie == DHCP_MagicCookie )
      {
        while ( UDPDataSize )
        {
          Option = * DataBuffer ++ ; UDPDataSize -- ;
          if ( Option == DHCP_PadCode )
          {
            continue ;
          } /* if */ ;
          if ( Option == DHCP_EndCode )
          {
            break ;
          } /* if */ ;
          OptionSize = * DataBuffer ++ ; UDPDataSize -- ;
          if ( OptionSize > UDPDataSize )
          {
            break ;
          } /* if */ ;
          switch ( Option )
          {
            case DHCP_MessageTypeCode :
              if ( OptionSize == 1 )
              {
                MessageType = * DataBuffer ++ ; UDPDataSize -- ;
                OptionSize = 0 ;
              } /* if */ ;
              break ;
            case DHCP_ServerIDCode :
              if ( OptionSize == 4 )
              {
                Server_ID = LocGetULONG ( DataBuffer ) ;
                DataBuffer += 4 ; UDPDataSize -= 4 ;
                OptionSize = 0 ;
              } /* if */ ;
              break ;
            case DHCP_SubNetMaskCode :
              if ( OptionSize == 4 )
              {
                SubNet = LocGetULONG ( DataBuffer ) ;
                DataBuffer += 4 ; UDPDataSize -= 4 ;
                OptionSize = 0 ;
              } /* if */ ;
              break ;
            case DHCP_RouterCode :
              if ( OptionSize == 4 )
              {
                GateWay = LocGetULONG ( DataBuffer ) ;
                DataBuffer += 4 ; UDPDataSize -= 4 ;
                OptionSize = 0 ;
              } /* if */ ;
              break ;
            case DHCP_LeaseTimeCode :
              if ( OptionSize == 4 )
              {
                Duration = LocGetULONG ( DataBuffer ) ;
                DataBuffer += 4 ; UDPDataSize -= 4 ;
                OptionSize = 0 ;
              } /* if */ ;
              break ;
          } /* switch */ ;
          if ( OptionSize )
          {
            DataBuffer += OptionSize ;
            UDPDataSize -= OptionSize ;
          } /* if */ ;
        } /* while */ ;
        UDPLayerInstance = UDPSocket -> UDPLayerInstanceLink ;
        IPLayerInstance = UDPLayerInstance -> IPLayerInstance ;
        switch ( MessageType )
        {
          case DHCP_DHCPOFFER :
            if ( DHCPInstance -> Command == DHCP_DHCPDISCOVER )
            {
              if ( DHCPInstance -> TimeOutService )
              {
                DetachTimer ( DHCPInstance -> TimeOutService ) ;
                DHCPInstance -> TimeOutService = NIL ;
              } /* if */ ;
              DHCPInstance -> Command = DHCP_DHCPREQUEST ;
              DHCPInstance -> Retry = DHCP_REQUEST_RETRIES ;
              DHCPInstance -> Client_IP = ClientIP ;
              DHCPInstance -> Server_ID = Server_ID ;
              Reply = TRUE ;
            } /* if */ ;
            break ;
          case DHCP_DHCPACK :
            if ( DHCPInstance -> Command == DHCP_DHCPREQUEST )
            {
              if ( DHCPInstance -> TimeOutService )
              {
                DetachTimer ( DHCPInstance -> TimeOutService ) ;
                DHCPInstance -> TimeOutService = NIL ;
              } /* if */ ;
              SetIPAdresses ( IPLayerInstance , ClientIP , SubNet , GateWay ) ;
              if ( DHCPInstance -> UserDCHPDataNotifyService )
                DHCPInstance -> UserDCHPDataNotifyService ( UDPSocket , ClientIP , SubNet , GateWay ) ;
              DHCPInstance -> Command = 0;
              if ( Duration != DHCP_INFINITE_DURATION )
              {
                if ( DHCPInstance -> TimeOutService )
                {
                  DetachTimer ( DHCPInstance -> TimeOutService ) ;
                  DHCPInstance -> TimeOutService = NIL ;
                } /* if */ ;
                DHCPInstance -> TimeOutService = AttachTimer ( ( 3 * TickPerSecond * Duration / 4 ) , 
                                                               DHCPLeaveTimeOutNotify , ( void * ) DHCPInstance ) ;
              } /* if */ ;
            } /* if */ ;
            break ;
          case DHCP_DHCPNACK :
            if ( DHCPInstance -> Command == DHCP_DHCPREQUEST )
            {
              if ( DHCPInstance -> TimeOutService )
              {
                DetachTimer ( DHCPInstance -> TimeOutService ) ;
                DHCPInstance -> TimeOutService = NIL ;
              } /* if */ ;
              SetIPAdresses ( IPLayerInstance , 0 , 0 , 0 ) ;
              DHCPInstance -> Command = DHCP_DHCPDISCOVER ;
              DHCPInstance -> XID ++ ;
              Reply = TRUE ;
            } /* if */ ;
            break ;
        } /* switch */ ;
        if ( Reply )
        {
          DHCPInstance -> TimeOut = DHCP_TimeOut ;
          SendDHCPRequest ( DHCPInstance ) ;
        } /* if */ ;
      } /* if */ ;
    } /* if */ ;
  } /* if */ ;
} /* DHCPDataRecvNotify */


static void DHCPTimeOutNotify ( void * CallParam )
{
  PtrDHCPLayerInstanceRecT DHCPInstance ;
  /*-------------------------------------------------------------------------*/
  DHCPInstance = ( PtrDHCPLayerInstanceRecT ) CallParam ;
  DetachTimer ( DHCPInstance -> TimeOutService ) ;
  DHCPInstance -> TimeOutService = NIL ;
  DHCPInstance -> Command = 0 ;
  DHCPClientRenew ( DHCPInstance ) ;
} /* DHCPTimeOutNotify */


static USHORT StrLen ( UCHAR * Str )
{
  UCHAR * SaveStr = Str ;
  /*-------------------------------------------------------------------------*/
  while ( * SaveStr ++ )
    ;
  return ( SaveStr - Str - 1 ) ;
} /* StrLen */


static USHORT SendDHCPRequest ( PtrDHCPLayerInstanceRecT DHCPInstance )
{
  PtrUDPLayerInstanceRecT UDPLayerInstance ;
  PtrIPLayerInstanceRecT IPLayerInstance ;
  EMACInstanceRecT * InterfInstance ;
  PtrOSFrameRecT OutFrame ;
  USHORT LocalDeviceNameSize ;
  USHORT Status ;
  /*-------------------------------------------------------------------------*/
  UDPLayerInstance = DHCPInstance -> UDPLayerInstance ;
  IPLayerInstance = UDPLayerInstance -> IPLayerInstance ;
  InterfInstance = IPLayerInstance -> InterfaceInstance ;
  OutFrame = InterfInstance -> InterfaceAllocFrameMemory ( sizeof ( OSFrameRecT ) ) ;
  if ( ! OutFrame )
    return ( UDP_SendFailed ) ;
  AddByteField ( OutFrame , DHCP_BootRequest ) ;              /* operacja */
  AddByteField ( OutFrame , EthHardwType ) ;                  /* typ sprzetu */
  AddByteField ( OutFrame , EthMACAddressSize ) ;             /* dlug. adresu sprzetowego */
  AddByteField ( OutFrame , 0 ) ;                             /* liczba skokow */
  AddULONGSwapField ( OutFrame , DHCPInstance -> XID ) ;      /* XID*/
  AddUSHORTSwapField ( OutFrame , 0 ) ;                       /* ilosc sekund */
  AddUSHORTSwapField ( OutFrame , 0x8000 ) ;                  /* flagi */
  AddULONGSwapField ( OutFrame , 0 ) ;                        /* adres IP klienta */
  AddULONGSwapField ( OutFrame , 0 ) ;                        /* przydzielony adres IP klienta */
  AddULONGSwapField ( OutFrame , 0 ) ;                        /* adres IP serwera */
  AddULONGSwapField ( OutFrame , 0 ) ;                        /* adres IP bramki */
  AddByteArrayField ( OutFrame , InterfInstance -> InterfaceMACAddress , EthMACAddressSize ) ;
  AddDuplByteField ( OutFrame , 0 , 16 - EthMACAddressSize ) ;/* adres sprztowy klienta (16 oktetw) */
  AddDuplByteField ( OutFrame , 0 , DHCP_SERVERNAME_SIZE ) ;  /* nazwa serwera (64 oktety) */
  AddDuplByteField ( OutFrame , 0 , DHCP_STARTFILE_SIZE ) ;   /* plik startowy (128 oktetw) */
                                                              /* options: */
  AddULONGSwapField ( OutFrame , DHCP_MagicCookie ) ;         /* DHCP_MagicCookie */
  AddByteField ( OutFrame , DHCP_MessageTypeCode ) ;          /*   */
  AddByteField ( OutFrame , 1 ) ;                             /*   */
  AddByteField ( OutFrame , DHCPInstance -> Command ) ;       /*   */
  AddByteField ( OutFrame , DHCP_HostNameCode ) ;
  LocalDeviceNameSize = StrLen ( DHCPInstance -> DeviceName ) ;
  AddByteField ( OutFrame , LocalDeviceNameSize ) ;
  AddByteArrayField ( OutFrame , DHCPInstance -> DeviceName , LocalDeviceNameSize ) ;
  if ( DHCPInstance -> Client_IP )
  {
    AddByteField ( OutFrame , DHCP_RequestedIPCode ) ;
    AddByteField ( OutFrame , 4 ) ;
    AddULONGSwapField ( OutFrame , DHCPInstance -> Client_IP ) ;
  } /* if */ ;
  if ( DHCPInstance -> Command == DHCP_DHCPREQUEST)
  {
    AddByteField ( OutFrame , DHCP_ServerIDCode ) ;
    AddByteField ( OutFrame , 4 ) ;
    AddULONGSwapField ( OutFrame , DHCPInstance -> Server_ID ) ;
  } /* if */ ;
  AddByteField ( OutFrame , DHCP_ParameterRequestListeCode ) ;
  AddByteField ( OutFrame , 2 ) ;
  AddByteField ( OutFrame , DHCP_SubNetMaskCode ) ;
  AddByteField ( OutFrame , DHCP_RouterCode ) ;
  AddByteField ( OutFrame , DHCP_EndCode ) ;
  if ( OutFrame -> Length < 300 )
  {
    AddDuplByteField ( OutFrame , 0 , 300 - OutFrame -> Length ) ;
  } /* if */ ;
traceDHCP(OutFrame -> Data , OutFrame -> Length);
UART0SendString ( ( UCHAR * ) "Powyzszy pakiet DHCP wyslano zakapsulowany jako UDP\r\n" ) ;
  Status = UDPSendData ( DHCPInstance -> DHCPSocket , 0xFFFFFFFF , DHCPServerPortNo ,
                         OutFrame -> Data , OutFrame -> Length ) ;
  InterfInstance -> InterfaceDeallocFrameMemory ( OutFrame ) ;
  if ( DHCPInstance -> TimeOutService )
  {
    DetachTimer ( DHCPInstance -> TimeOutService ) ;
  } /* if */ ;
  DHCPInstance -> TimeOutService = AttachTimer ( DHCPInstance -> TimeOut , DHCPTimeOutNotify ,
                                                 ( void * ) DHCPInstance ) ;
  return ( Status ) ;
} /* SendDHCPRequest */


void DHCPClientRenew ( PtrDHCPLayerInstanceRecT DHCPInstance )
{
  /*-------------------------------------------------------------------------*/
  if ( ! DHCPInstance -> Command )
  {
    DHCPInstance -> XID ++ ;
    DHCPInstance -> Command = DHCP_DHCPDISCOVER ;
    DHCPInstance -> TimeOut = DHCP_TimeOut ;
    if ( DHCPInstance -> TimeOutService )
      DetachTimer ( DHCPInstance -> TimeOutService ) ;
    DHCPInstance -> TimeOutService = AttachTimer ( DHCPInstance -> TimeOut , DHCPTimeOutNotify ,
                                                   ( void * ) DHCPInstance ) ;
    SendDHCPRequest ( DHCPInstance ) ;
  } /* if */ ;
} /* DHCPClientRenew */


USHORT StartDHCPService ( PtrDHCPLayerInstanceRecT    DHCPInstance ,
                          RecVDHCPDataNotifyServProcT DataNotifyService )
{
  PtrUDPLayerInstanceRecT UDPLayerInstance ;
  USHORT Status ;
  /*-------------------------------------------------------------------------*/
  if ( ! DHCPInstance -> DHCPSocket )
  {
    UDPLayerInstance = DHCPInstance -> UDPLayerInstance ;
    DHCPInstance -> UserDCHPDataNotifyService = DataNotifyService ;
    DHCPInstance -> DHCPSocket = CreateUDPSocket ( UDPLayerInstance ) ;
    if ( ! DHCPInstance -> DHCPSocket )
      return ( FALSE ) ;
    Status = BindUDPSocket ( DHCPInstance -> DHCPSocket , DHCPClientPortNo ) ;
    if ( Status != UDP_BindPortOK )
      return ( FALSE ) ;
    SetUDPListen ( DHCPInstance -> DHCPSocket , DHCPDataRecvNotify ) ;
  } /* if */ ;
UART0SendString ( ( UCHAR * ) "DHCPClientRenew ( DHCPInstance ) ; -> start negocjacji DHCP\r\n" ) ;
  DHCPClientRenew ( DHCPInstance ) ;
  return ( TRUE ) ;
} /* StartDHCPService */


void StopDHCPService ( PtrDHCPLayerInstanceRecT DHCPInstance )
{
  /*-------------------------------------------------------------------------*/
  DHCPInstance -> UserDCHPDataNotifyService = NIL ;
  if ( DHCPInstance -> DHCPSocket )
    DeallocateUDPSocket ( DHCPInstance -> DHCPSocket ) ;
  DHCPInstance -> DHCPSocket = NIL ;
  if ( DHCPInstance -> TimeOutService )
  {
    DetachTimer ( DHCPInstance -> TimeOutService ) ;
    DHCPInstance -> TimeOutService = NIL ;
  } /* if */ ;
} /* StopDHCPService */


PtrDHCPLayerInstanceRecT AllocateDHCPClientInstance ( PtrUDPLayerInstanceRecT UDPLayerInstance ,
                                                      UCHAR *                 DeviceIdentifier )
{
  PtrDHCPLayerInstanceRecT DHCPInstance ;
  USHORT Loop ;
  USHORT LocalDeviceNameSize ;
  /*-------------------------------------------------------------------------*/
  DHCPInstance = & LocalDHCPLayerInstance ;
  DHCPInstance -> UDPLayerInstance = UDPLayerInstance ;
  DHCPInstance -> DHCPSocket = NIL ;
  if ( ! DeviceIdentifier )
  {
    LocalDeviceNameSize = 0 ;
  } /* if ... */
  else
  {
    LocalDeviceNameSize = StrLen ( DeviceIdentifier ) ;
  } /* if ... else */ ;
  for ( Loop = 0 ; Loop < DeviceNameSize ; Loop ++ )
    DHCPInstance -> DeviceName [ Loop ] = 0 ;
  if ( LocalDeviceNameSize >= DeviceNameSize )
    LocalDeviceNameSize = DeviceNameSize - 1 ;
  if ( LocalDeviceNameSize )
  {
    for ( Loop = 0 ; Loop < LocalDeviceNameSize ; Loop ++ )
      DHCPInstance -> DeviceName [ Loop ] = * DeviceIdentifier ++ ;
  } /* if */ ;
  DHCPInstance -> Command = 0 ;
  DHCPInstance -> Retry = 0 ;
  DHCPInstance -> Client_IP = 0 ;
  DHCPInstance -> Server_ID = 0 ;
  DHCPInstance -> XID = MagicID ( UDPLayerInstance ) ;
  DHCPInstance -> TimeOut = 0 ;
  DHCPInstance -> TimeOutService = NIL ;
  return ( DHCPInstance ) ;
} /* AllocateDHCPClientInstance */


